/*
 * SocketWrench 10
 * Copyright 2020 Catalyst Development Corporation
 * All rights reserved
 *
 * This module contains the confidential and proprietary information
 * of Catalyst Development Corporation and may not be redistributed.
 *
 * This file is licensed to you pursuant to the terms of the
 * product license agreement included with the original software,
 * and is protected by copyright law and international treaties.
 * Unauthorized reproduction or distribution may result in severe
 * criminal penalties.
 */

#include <stdio.h>
#include "cswsock10.h"

#define SERVER_PORT		7		// Local port number for server
#define SERVER_CLIENTS	1000	// Maximum number of client connections

VOID ShowError(LPCTSTR lpszFunction);
VOID CALLBACK EventHandler(SOCKET hSocket, UINT nEventId, DWORD dwError, DWORD_PTR dwParam);

int _tmain(int argc, _TCHAR* argv[])
{
	SOCKET hServer = INVALID_SOCKET;

	if (!InetInitialize(NULL, NULL))
	{
		ShowError(_T("InetInitialize"));
		return 0;
	}

	// Start the server
	hServer = InetServerStart(NULL,
		SERVER_PORT,
		INET_BACKLOG,
		SERVER_CLIENTS,
		INET_TIMEOUT,
		INET_PRIORITY_DEFAULT,
		INET_OPTION_REUSEADDRESS,
		EventHandler,
		0,
		NULL);


	if (hServer == INVALID_SOCKET)
	{
		ShowError(_T("InetServerStart"));
		return 0;
	}

	_tprintf(_T("Server started on port %d\n"), SERVER_PORT);

	while (TRUE)
	{
		TCHAR szCommand[128];
		SIZE_T cchCommand;

		_tprintf(_T("> "));
		fflush(stdout);

		// Read a command from standard input
		if (_fgetts(szCommand, 128, stdin) == NULL)
			break;

		cchCommand = _tcslen(szCommand);

		// Strip any trailing newline from the command string
		if (cchCommand > 0 && szCommand[cchCommand - 1] == '\n')
			szCommand[cchCommand - 1] = '\0';

		if (_tcsicmp(szCommand, _T("BROADCAST")) == 0)
		{
			// Broadcast a test message to all clients connected to the server
			LPCSTR lpszMessage = "\r\nThis is a test broadcast message\r\n";

			if (InetServerBroadcast(hServer, (LPBYTE)lpszMessage, lstrlenA(lpszMessage)) == INET_ERROR)
				ShowError(_T("InetBroadcast"));
		}
		else if (_tcsicmp(szCommand, _T("RESUME")) == 0)
		{
			// The server will resume accepting connections
			if (InetServerResume(hServer))
				_tprintf(_T("Server resumed, accepting client connections\n"));
			else
				ShowError(_T("InetServerResume"));
		}
		else if (_tcsicmp(szCommand, _T("RESTART")) == 0)
		{
			// Restart the server, with the new server socket handle being
			// returned by the function
			hServer = InetServerRestart(hServer);

			if (hServer != INVALID_SOCKET)
				_tprintf(_T("Server restarted, accepting client connections\n"));
			else
				ShowError(_T("InetServerRestart"));
		}
		else if (_tcsicmp(szCommand, _T("STATUS")) == 0)
		{
			// Display the number of clients connected to the server and
			// the current status of the server
			INT nClients = InetEnumServerClients(hServer, NULL, 0);
			_tprintf(_T("There are %d clients connected to the server\n"), nClients);

			switch (InetGetServerStatus(hServer))
			{
			case INET_SERVER_LISTENING:
				_tprintf(_T("The server is listening for client connections\n"));
				break;

			case INET_SERVER_SUSPENDED:
				_tprintf(_T("The server is not listening for client connections\n"));
				break;

			default:
				_tprintf(_T("The server is in an unexpected state\n"));
				break;
			}
		}
		else if (_tcsicmp(szCommand, _T("STOP")) == 0 || _tcsicmp(szCommand, _T("QUIT")) == 0)
		{
			// Exit the loop which will stop the server		
			_tprintf(_T("Stopping server\n"));
			break;
		}
		else if (_tcsicmp(szCommand, _T("SUSPEND")) == 0)
		{
			// The server will stop accepting client connections
			if (InetServerSuspend(hServer))
				_tprintf(_T("Server suspended, no longer accepting client connections\n"));
			else
				ShowError(_T("InetServerSuspend"));
		}
		else
		{
			_tprintf(_T("Unrecognized command\n"));
		}
	}

	// Stop the server and uninitialize the library
	InetServerStop(hServer);
	InetUninitialize();
	return 0;
}

// Display an error message
VOID ShowError(LPCTSTR lpszFunction)
{
	DWORD dwError;
	TCHAR szError[256] = _T("Unknown error");

	dwError = InetGetLastError();

	if (dwError > 0)
		InetGetErrorString(dwError, szError, 256);

	_ftprintf(stderr, _T("%s: %s (0x%08lX)\n"), lpszFunction, szError, dwError);
}

VOID CALLBACK EventHandler(SOCKET hSocket, UINT nEventId, DWORD dwError, DWORD_PTR dwParam)
{
	switch (nEventId)
	{
	case INET_EVENT_READ:
		if (dwError == 0)
		{
			BYTE cBuffer[1024];
			INT nBytesRead;
			INT nBytesWritten;

			// Read data sent from the client to the server
			nBytesRead = InetRead(hSocket, cBuffer, sizeof(cBuffer));

			// Send that data back to the client
			if (nBytesRead > 0)
				nBytesWritten = InetWrite(hSocket, cBuffer, nBytesRead);
		}
		break;
	}
}
